import java.util.Hashtable;
import java.util.Map;

import org.json.JSONObject;

import redis.clients.jedis.Jedis;


/**
 * Users spawn threads that subscribe to Redis channels and wait for incoming messages.
 * This keeps track of user state and whatnot.
 * Also implements command logic.
 * 
 * Commands are:
 * 		me			/me
 * 		join		/join
 * 		whois		/whois
 * 		tell		/tell
 * 		delete		/delete
 * 		leave		/leave
 *		sendMessage /chat 
 *
 * @author Stephen Smithbower 2013
 */
public class User 
{
	//////////////////////////////////////////////////////
	// INSTANCE DATA									//
	//////////////////////////////////////////////////////
	/**
	 * The user's unique username.
	 */
	public String username;
	
	/**
	 * Host Redis IP address.
	 */
	public String host;
	
	/**
	 * User connection used to publish messages on. Messages broadcasted on all
	 * channels will use this connection. Separate connections are used per-channel.
	 */
	public Jedis conn;
	
	/**
	 * Tracks all channels we're subscribed to so we can close them up properly.
	 */
	public Hashtable<String, Channel> channels = new Hashtable<String, Channel>();
	
	/**
	 * Recipient of events when a message is received.
	 */
	private IMessageEventListener messageListener;
	
	
	
	
	//////////////////////////////////////////////////////
	// METHODS											//
	//////////////////////////////////////////////////////
	/**
	 * Creates a new User connection to Redis.
	 * 
	 * @param host The host Redis IP address.
	 * @param messageListener The recipient of events when a message is received.
	 */
	public User(String host, IMessageEventListener messageListener)
	{
		this.messageListener = messageListener;
		
		this.username = ""; //Blank, because we don't have a name yet. Should be set by whois.
		
		this.host = host;
		this.conn = new Jedis(host, 50030);
	}
	

	/**
	 * Has the user join a channel.
	 * 
	 * @param channel The name of the channel to join.
	 */
	public void join(String channel)
	{
		if (this.username.equals(""))
			return; //TODO: Throw exception.
		
		//Make sure we haven't already joined this channel.
		if (this.channels.containsKey("channel:" + channel))
			return; //TODO: Throw exception.
		
		//Update channel count.
		long channelCount = this.conn.sadd("channels:" + this.username, "channel:" + channel);
		
		//Start listening to the channel.
		this.channels.put("channel:" + channel, new Channel(this.host, "channel:" + channel, this.messageListener, this));
		
		//Notify users that we've joined the channel.
		if (channelCount > 0)
		{
			JSONObject json = new JSONObject();
			json.put("name", "SERVER");
			json.put("message", this.username + " has joined channel: " + channel);
			
			//conn.publish("channel:" + name, "CANCEL"); //What?
			this.conn.publish("channel:" + channel, json.toString());
		}
		
		//Refresh user expiration countdown.
		refreshUserExpiration();
	}
	
	
	/**
	 * Has the user identify itself to the server.
	 * 
	 * @param name The name of the user. If the user already has a different name,
	 * 			   and this name is available, we delete the old name and use this
	 * 			   one.
	 * @param age The age of the user.
	 * @param sex The sex of the user (male/female).
	 * @param location The user's location (ASL??)
	 */
	public void me(String name, int age, String sex, String location)
	{
		//Check to see if the user already exists.
		if (this.conn.exists("user:" + name))
			return;//TODO: Throw an exception "[me] Error: That user already exists!";
		
		//If we're renaming ourselves, delete the old name.
		if (!this.username.equals(""))
		{
			this.conn.del("user:" + this.username);
			this.conn.del("channels:" + this.username);
			
			//Stop listening on channel:all, and old username.
			Channel chanAll = this.channels.get("channel:all");
			Channel chanOldname = this.channels.get("channel:" + this.username);
			
			chanAll.stop();
			chanOldname.stop();
			
			this.channels.remove("channel:all");
			this.channels.remove("channel:" + this.username);
		}
		
		//Ok, add ourselves.
		this.username = name;
		Map<String, String> vals = new Hashtable<String, String>();
		vals.put("name", this.username);
		vals.put("age", Integer.toString(age));
		vals.put("sex", sex);
		vals.put("location", location);
		this.conn.hmset("user:" + this.username, vals);
		
		//Delete the user if they sit idle for 3 minutes.
		refreshUserExpiration();
		
		//Update channel listing.
		this.conn.sadd("channels:" + this.username, "channel:" + this. username, "channel:all");
		
		
		//Subscribe to channels.
		this.channels.put("channel:all", new Channel(this.host, "channel:all", this.messageListener, this));
		this.channels.put("channel:" + this.username, new Channel(this.host, "channel:" + this.username, this.messageListener, this));
		
		
		//Notify everyone that we've joined the server.
		JSONObject json = new JSONObject();
		json.put("name", "SERVER");
		json.put("message", this.username + " has joined the server");
		this.conn.publish("channel:all", json.toString());
	}
	
	
	/**
	 * Gets some data on a given user.
	 * 
	 * @param user The name of the user to get data for.
	 * 
	 * @return A key-value pairing of data on the user.
	 */
	public Map<String, String> whois(String user)
	{
		if (this.username.equals(""))
			return null; //TODO: Throw exception.
		
		//Refresh user expiration countdown.
		refreshUserExpiration();
		
		//Get the Redis data structure.
		return this.conn.hgetAll("user:" + user);
	}
	
	
	/**
	 * Has the user leave a channel.
	 * 
	 * @param channel The name of the channel to leave.
	 */
	public void leave(String channel)
	{
		if (this.username.equals(""))
			return; //TODO: Throw exception.
		
		//Make sure we've joined this channel.
		if (!this.channels.containsKey("channel:" + channel))
			return; //TODO: Throw exception.
		
		//Make sure the channel isn't empty.
		long channelCount = this.conn.srem("channels:" + this.username, "channel:" + channel);
		
		if (channelCount > 0)
		{
			//Notify all users in the channel that we've left.
			JSONObject json = new JSONObject();
			json.put("name", "SERVER");
			json.put("message", this.username + " has left channel: " + channel);
			this.conn.publish("channel:" + channel, json.toString());
		}
		
		//Stop listening on this channel.
		Channel chanByeBye = this.channels.get("channel:" + channel);
		chanByeBye.stop();
		this.channels.remove("channel:" + channel);
		
		//Refresh user expiration countdown.
		refreshUserExpiration();
	}
	
	
	/**
	 * Broadcasts a message to a given channel.
	 * 
	 * @param channel The name of the channel to broadcast on. Defaults to all
	 * 				  if none is specified.
	 * @param message The message to send.
	 */
	public void sendMessage(String channel, String message)
	{
		if (this.username.equals(""))
			return; //TODO: Throw exception.
		
		if (channel.equals("")) //Default to "all"
			channel = "all";
		
		//Make sure we've joined this channel.
		if (!this.channels.containsKey("channel:" + channel))
			return; //TODO: Throw an exception.
		
		//Create the message.
		JSONObject json = new JSONObject();
		json.put("name", this.username);
		json.put("channel", channel);
		json.put("message", message);
		
		//Publish the message.
		this.conn.publish("channel:" + channel, json.toString());
		
		//Refresh user expiration countdown.
		refreshUserExpiration();
	}
	
	
	/**
	 * Drops the user off a cliff.
	 */
	public void delete()
	{
		if (this.username.equals(""))
			return; //TODO: Throw exception.
		
		this.conn.del("user:" + this.username);
		this.conn.del("channels:" + this.username);
		
		//Let everyone else on the server know that we've left.
		JSONObject json = new JSONObject();
		json.put("name", this.username);
		json.put("message", this.username + " has left the server");
		this.conn.publish("channel:all", json.toString());
		
		//Stop listening on all channels.
		for(String chan : this.channels.keySet())
			this.channels.get(chan).stop();
		
		this.channels.clear();
		
	}
	
	
	/**
	 * Sends a private message to a given user on their private channel.
	 * 
	 * @param user The name of the user to send the message to.
	 * @param message The message to send to the other user.
	 */
	public void tell(String user, String message)
	{
		if (this.username.equals(""))
			return; //TODO: Throw exception.
		
		//Create the message packet.
		JSONObject json = new JSONObject();
		json.put("name", this.username);
		json.put("channel", user);
		json.put("message", message);
		
		//Whisper the user on their private channel.
		this.conn.publish("channel:" + user, json.toString());
		
		//Refresh user expiration countdown.		
		refreshUserExpiration();
	}
	
	
	/**
	 * Resets user timeouts to 3 minutes, after which they will be kicked.
	 */
	private void refreshUserExpiration()
	{
		this.conn.expire("user:" + this.username, 60 * 3);
		this.conn.expire("channels:" + this.username, 60 * 3);
	}
}
